/*
  Frequ synth for for FM828
  Denys VK3ZYZ (with help from Google and others)
  Fixes by Ray and Josh 20190728
    CTCSS as per https://www.amateurradio.com.au/repeaters
  Add selectable CTCSS and frequencies 20230315
    Still to do:
      indicate the freq change mode.
      selectable sim/duplex with band limits test
      Save selections to EEPROM for next time  
*/

#define Version "FM828A_VFO_20230320"

#include "si5351.h"
#include "Wire.h"
#include "Adafruit_GFX.h"
#include "Adafruit_SSD1306.h"
#define SCREEN_WIDTH 128 // OLED display width, in pixels
#define SCREEN_HEIGHT 32 // OLED display height, in pixels

// Declaration for an SSD1306 display connected to I2C (SDA, SCL pins)
#define OLED_RESET     -1 // Reset pin # (or -1 if sharing Arduino reset pin)
Adafruit_SSD1306 display(SCREEN_WIDTH, SCREEN_HEIGHT, &Wire, OLED_RESET);

// synthesiser default settings on start up.
unsigned long RXgenfrequency  = 146000000;
unsigned long TXgenfrequency  = 146000000;
unsigned long IFoffset        = 10700000;    // IF offset
unsigned long TXmultiplier    = 8;           // transmitter frequency multiplier
unsigned long RXmultiplier    = 6;           // receiver frequency multiplier. Should be 3 but there is a problem with that.
#define       MinFreq           145000       // set the band extremes (in khz)
#define       MaxFreq           147975
float         DisplayFreq     = 146.000;
#define       forward           HIGH
int           mode            = forward;     // HIGH = normal, LOW = REV
#define       RX                LOW
int           TRXmode         = RX;          // HIGH = TX, LOW = RX
int           channel         = 2;           // convert frequence to channel number
int encoder;                                 // encoder direction: CCW=-1; NC=0; CW=1
int timer = 100;           // sub tone setting. The higher the number, the slower the timing.
int  i ;
unsigned long temp;
int temp2;

int32_t cal_factor = 32400; //  <<<--- found from Si5351 calibrator sketch.

Si5351 clockgen;                             // initalize clockgen

// encoder definitions
#define ENCODER_A     A0
#define ENCODER_B     A1
#define spare         A2           // not used yet
#define ModeSwitch    A3           // extra mode selector
// A4 = SDA
// A5 = SCL
#define AnalogIn1     A6           // not used yet
#define AnalogIn2     A7           // not used yet
 
#define TRX            2           // input HIGH = TX, LOW = RX.
#define Input2         3           // not used yet
#define Output1        4           // not used yet
#define REV_SWITCH     5           // Encoder Button  repeater reverse
#define spareSW        6           // not used yet

/* ********************* PortB0-PB5 used to generate CTCSS tones via R2R nerwork.
#define spare1         8           // Sub Tone R2R b0 PB0
#define spare2         9           // Sub Tone R2R b1 PB1
#define spare3        10           // Sub Tone R2R b2 PB2
#define spare4        11           // Sub Tone R2R b3 PB3
#define spare5        12           // Sub Tone R2R b4 PB4
#define spare6        13           // Sub Tone R2R b5 PB5
*/

#define ENCODER_INPUT_PORT PINC
#define On  1
#define Off 0
#define RefFrequ      500000        // 16Mhz / 32 steps for CTCSS sine wave generator
int     SineDivider = 4065;         // default sine wave divider

  
long int RXFrequs []      // Edit this for RX frequencies wanted.
{ 146000,
  146625,
  146650,
  146675,
  146700,
  146725,
  146750,
  146775,
  146800,
  146825,
  146850,
  146875,
  146900,
  146925,
  146950,
  146975,
  147000,
  146425,
  146450,
  146475,
  146500,
  146525,
  146550,
  146575,
  146600,
  147025,
  147050,
  147075,
  147100,
  147125,

  147150,
  147175,
  147200,
  147225,
  147250,
  147275,
  147300,
  147325,
  147350,
  147375,

};

int TXFrequs []      // Edit this for TX frequencies wanted. Added to RX frequs.
{ 0,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  -600,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  600,
  600,
  600,
  600,
  600,

  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,
  600,

};

// Changing SubTone to int lets more channels be defined.
int SubTone []      // Edit this for CTCSS frequencies wanted.(freq x 10)
{ 0,      // "0 ="  chan 1
  0,
  19,  //  chan 3
  0,
  10,   //  chan 5
  0,
  10,   //  chan 7
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  10,   // chan 15
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  0,
  19,    // chan 26
  0,
  0,
  10,   // chan 29
  10,   // chan 30

  0,
  0,
  0,
  0,
  0,
  10,   // chan 36
  0,
  0,
  0,
  10,   // chan 40

};

float CTCSS = 0;
int   CTCSS_Index = 0;
int   CTCSS_Tones []
//   0     1     2     3     4     5     6     7     8     9
{    0,  670,  693,  719,  744,  770,  797,  825,  854,  885,   //  +0
   915,  948,  974, 1000, 1035, 1072, 1109, 1148, 1188, 1230,   // +10
  1273, 1318, 1365, 1413, 1462, 1500, 1514, 1567, 1598, 1622,   // +20
  1655, 1679, 1713, 1773, 1799, 1835, 1862, 1899, 1928, 1966,   // +30
  1995, 2035, 2065, 2107, 2181, 2257, 2291, 2336, 2418, 2503,   // +40
  2541,                                                         // +50
};

int sinetable[] // one cycle sine wave via R2R network.
{ 0,1,2,3,5,7,10,13,16,19,22,25,27,28,29,30,31,30,29,28,27,25,22,19,16,13,10,7,5,3,2,1, };


void setup() {
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C); // Address 0x3C for 128x32
  display.clearDisplay();
  display.display();                  // Update screen
  display.setTextSize(1);             // Normal 1:1 pixel scale
  display.setTextColor(WHITE,BLACK);  // Draw white text
  display.setCursor(0,0);             // Start at top-left corner
  display.println(F(Version));
  display.display(); // Update screen
  delay(1000);
  display.clearDisplay();
  display.display();                  // Update screen
  delay(100);
  display.setCursor(0,0);
  display.print("Chan                  ");
  display.display(); // Update screen
  clockgen.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);
  clockgen.set_correction(cal_factor, SI5351_PLL_INPUT_XO); // <<<<<<<this is found from Si5351 calibrator code.

  // configure rotary encoder & push-button
  // do not use INPUT_PULLUP if encoder has built in pulups
  pinMode(ENCODER_A, INPUT);     // encoder leads
  pinMode(ENCODER_B, INPUT);
  pinMode(REV_SWITCH, INPUT);    // Hi = nomal, LOW =  Reverse
  pinMode(TRX, INPUT);     // Hi = TX, LOW =  RX
  pinMode(Input2, INPUT_PULLUP);
  pinMode(Output1, OUTPUT);
  DDRB  = B111111;                // sets Arduino pins 7 to 12 as outputs.
  PORTB = B100000;                // R2R out at rest to half level.
  digitalWrite(Output1, HIGH);    // output1 off

  encoder = read_encoder();       // returns +1/0/-1 (cw/nc/ccw)
  mode=( digitalRead(REV_SWITCH));
  TRXmode=(digitalRead(TRX));
  set_synth(mode, TRXmode);
}

/******************************
  main loop
*******************************/
void loop() {

  if ( digitalRead(REV_SWITCH) != mode)           // has the Reverse switch changed?
    { mode=( digitalRead(REV_SWITCH));            // Yes, set the mode to suit.
    set_synth(mode, TRXmode);                     // set the PLL
    }
  if (digitalRead(TRX) !=TRXmode)                 // has the TX/RX mode changed?
    { TRXmode=(digitalRead(TRX));                 // Yes, set the mode to suit.
      set_synth(mode, TRXmode);                   // set the PLL
    }
  if (TRXmode==RX)                                // check for frequency chanegs if RX only.
    { encoder = read_encoder();                   // returns +1/0/-1 (cw/nc/ccw)
      if (encoder != 0)                           // has the encoder been moved?
        { channel = channel + encoder;            // Yes, inc/dec channel number
          channel = constrain(channel, 0, 39);    // max 40 channels.
          set_synth(mode, TRXmode);               // set the PLL
        }
          temp2 =  analogRead(ModeSwitch);        // 2 readings the same? NOTE! these numbers
      if (temp2 == analogRead(ModeSwitch))        // will depend on Rs used on ModeSwitch.
      { if      (temp2 >= 550)   // 550 600       // both buttons pushed (on my IC22S)  
         { // something();      
         }
        else  if (temp2 >= 400)  // 400 504       // Left button pushed (on my IC22S)
         { AdjustCTCSS();      
         }
        else if (temp2 >= 200)   // 200 320       // Right button pushed (on my IC22S)
        { AdjustChannel();  
        }
      } 
  }                                               // end of RX mode adjustments
}

/******************************
  read encoder direction
*******************************/
int read_encoder() {
  // definitions
#define DELAY 100                               // number of polling-loops   
  static unsigned short integrator;               // encoder debounce

  /* the array values are determined by combining the previous and current A1,A0
    Gray Code bit-patterns to form an array-index. +1 is assigned to all clockwise
    (CW) indexes; -1 is assigned to all counter-clockwise (CCW) indexes; and 0 is
    assigned to the remaining positions to indicate no movement*/
  static char encoder[] = {0, -1, 1, 0, 1, 0, 0, -1, -1, 0, 0, 1, 0, 1, -1, 0}; // SWAPPED PHASES
  static unsigned char encoder_index = 0;
  static long count = 0;                         // CW rotation = +1; CCW rotation = -1
  static long this_reading;
  static long last_reading;
  char value;                                    // return value

  /* check for encoder rotation. The absolute "count" value changes as the encoder rotates.
    Typically there are four states (counts) per click, but often more due to contact bounce.
    Bounce doesn't matter so long as the "count" decreases" for CCW rotation; remains constant
    when there is no rotation; and increases with CW rotation. Bounce can be completely
    eliminated if we look for the "detent" code and integrate as shown below. */
  encoder_index <<= 2;                           // shift previous bit-pattern left two places
  encoder_index |= (ENCODER_INPUT_PORT & 0x03);  // get current bit-pattern and combine with previous
  count += encoder[( encoder_index & 0x0f )];

  // debounce encoder
  if ((ENCODER_INPUT_PORT & 0x03) == 3) {        // encoder on an indent (3 decimal == 00000011 binary)
    if (integrator < DELAY) {
      integrator++;
    }
  } else {                                       // encoder not resting on indent
    if (integrator > 0) {
      integrator--;
    }
  }

  // prepare the "return" value
  if (integrator >= DELAY) {                     // encoder deemed to be stationary
    this_reading = count;
    if (this_reading < last_reading) value = -1;
    if (this_reading == last_reading) value = 0;
    if (this_reading > last_reading) value = 1;
    last_reading = this_reading;                 // both readings now hold current "count"
    integrator = 0;                              // reset integrator
  } else {
    value = 0;
  }
  return (value);
}

/*******************************
  set the Si5351 CLK0 frequency
 ******************************/
 
void set_synth( int frmode, int trx )       // set the display and PLL
{  
  display.setTextSize(1);                   // Normal 1:1 pixel scale
  display.setCursor(18,0);                  // write channel number to the LCD 60
  display.print("  ");                      // clear the old number
  display.setCursor(18,0);                  // write channel number to the LCD
  if (channel < 9)                          // "9" = 10 as array starts at 0
  { display.print(" ");                     // extra space for formalling a single digit.
  }
  display.print((channel + 1));             // display the channel number
  display.setCursor(34,0);                  // write mode to the LCD
  if (TXFrequs[channel]== 0)
  {display.print(" SIM "); }
    if (TXFrequs[channel] == 600)
  {display.print("+600"); }
    if (TXFrequs[channel] ==-600)
  {display.print("-600"); }
  ShowCTCSS();                             // display the CTCSS status
                                           // Select FWD or REV frequencies
  if (frmode == HIGH)                      // test for fwd or rev repeater...
  { RXgenfrequency = (RXFrequs[channel]);  // normal operation
    TXgenfrequency = ((RXFrequs[channel]) + (TXFrequs[channel]));
  }
  else
  { TXgenfrequency = (RXFrequs[channel]); // reverse repeater operation
    RXgenfrequency = ((RXFrequs[channel]) + (TXFrequs[channel]));
  }                                       // Then display them....
  if ( trx == RX)                         // is it = RX?
    { DisplayFreq = RXgenfrequency;       // Yes, display the RX frequency
      ShowFreq();
    }
   else                                   // No, it is TX.
    { display.setTextColor(BLACK,WHITE);  // Invert Frequ Display
      DisplayFreq = TXgenfrequency;       // display the TX frequency
      ShowFreq();
    }
  display.setTextColor(WHITE,BLACK);      // back to normal text
  display.display();                      // Update screen
   
// calculate the actual frequencies to set

  RXgenfrequency = (((RXgenfrequency * 1000) - IFoffset) / RXmultiplier);
  TXgenfrequency = ((TXgenfrequency  * 1000) / TXmultiplier);

// is it transmittig or receiving?

  if ( trx == RX) // =RX.
  { clockgen.output_enable(SI5351_CLK0, 1);                 // turn RX gen on
    clockgen.set_freq((RXgenfrequency * 100), SI5351_CLK0); // set the RX generator frequ,
    clockgen.output_enable(SI5351_CLK1, 0);                 // turn TX gen off
    TIMSK1 = 0;                                             // subtone off
  }
  else                            // =TX.
  { clockgen.output_enable(SI5351_CLK1, 1);                 // turn TX gen on
    clockgen.set_freq((TXgenfrequency * 100), SI5351_CLK1); // set the TX generator frequ,
    clockgen.output_enable(SI5351_CLK0, 0);                 // turn RX gen off
    if ((SubTone[channel]) != 0)
    { SineDivider = (RefFrequ/(float)(SubTone[channel]/10)); // Subtone in as Int = Float/10
    CTCSS_Tone();                                           // subtone on
    }
    else 
    {
    TIMSK1 = 0;                                             // subtone off
    }
  }
}

void CTCSS_Tone()                         // CTCSS tone generation setup
{ cli();                                  // interupsts off for setup
  TCCR1A = 0;                             // 
  TCCR1B = 0;                             //  no prescaler
                                          // set compare match register for 1hz increments
  OCR1A = SineDivider ;                   // = (16*10^6) / (1*1024) - 1 (must be <65536)
  TCCR1B |= (1 << WGM12);                 // turn on CTC mode
  TCCR1B |= (0 << CS12) | (0 << CS11) | (1 << CS10);    // Set CS10 no prescaler
  TIMSK1 |= (1 << OCIE1A);                // enable timer compare interrupt
  sei();                                  // interupsts on      
}

ISR (TIMER1_COMPA_vect)                   // generate the CTCSS tone
   { PORTB=(sinetable[i++]);              // write value from table to PORTB
     if(i>=32)                            // 32 step loop for next value
      { i=0;
      }
  }

void ShowCTCSS()                          // show the CTCSS frequency..
{ display.setTextSize(1);                 // Normal 1:1 pixel scale
  display.setCursor(70,0);                // set cursor
  CTCSS_Index=(SubTone[channel]);         // get the CTCSS number
  if(CTCSS_Index!=0)                      // if not zero,
    { display.print("tone");              // print "tone"
      CTCSS = ((CTCSS_Tones[CTCSS_Index])/10);  // then get the actual CTCSS freq in hz
      if(CTCSS<100)                       // is it less than 100hz?
        { display.print(" ");             // pad leading space if needed
        }  
      display.print(CTCSS,1);             // display the CTCSS frequ to 1 decimal place
    }
    else { display.print("         ");    // no CTCSS for this channel, so clear the CTCSS display.
         }
  display.display();                      // update the display.
}

void AdjustCTCSS()                        // modify the CTCSS number index for this channel.
  {   display.setTextColor(BLACK,WHITE);  // invert text to indicate edit mode
      ShowCTCSS();                        // display the current CTCSS. 
 while (analogRead(ModeSwitch) >= 400)    // while in edit CTCSS mode....
    { encoder = read_encoder();           // look at the encoder
     if (encoder != 0)                    // Has the encoder moved?
      { temp2 = (SubTone[channel]);        // Yes, get the current CTCSS number (0-50)
        temp2 = temp2 + encoder;          // add in te encoder
        temp2 = constrain(temp2, 0, 50);  // limit the number to the valid range       
        (SubTone[channel]) = temp2;       // put the new number back in te table
        ShowCTCSS();                      // then show the new CTCSS.
      } 
    }                                     // go back to do it again if mode still valid
      display.setTextColor(WHITE,BLACK);  // back to normal text
      ShowCTCSS();                        // show the final CTCSS in normal text, not inverted.
}

void ShowFreq()                           // display the current frequency for this channel
{ display.setCursor(1,10);                // position the cursor
  display.setTextSize(3);                 // and large text
  DisplayFreq = (DisplayFreq)/1000;       // convert the freq (in khz) to Mhz
  display.print(DisplayFreq, 3);          // display the frequ fo 3 decimal places.
  display.display();                      // Update screen
}

void ShowFreqAdj()                        // display the freqi for channel being edited
{ display.setCursor(1,10);                // position the cursor 
  display.setTextSize(3);                 // and large text
  display.print("14");                    // display the "14" part of the 14xxxx
  DisplayFreq = (DisplayFreq - 140000);   // remove the "14" part of the frequency
  DisplayFreq = (DisplayFreq)/1000;       // convert remainder to Mhz
  display.setTextColor(BLACK,WHITE);      // invert text
  display.print(DisplayFreq, 3);          // display the frequ fo 3 decimal places.
  display.display();                      // Update screen
  display.setTextColor(WHITE,BLACK);      // back to normal text
}

void AdjustChannel()                      // Adjust the channel frequency
{    DisplayFreq = (RXFrequs[channel]);   // get the current channel frequency
     ShowFreqAdj();                       // display the frequency

 while (analogRead(ModeSwitch) >= 300)    // while in adjust channel frequency mode
    { encoder = read_encoder();           // look at the encoder
     if (encoder != 0)                    // has it moved?
      { temp = (RXFrequs[channel]);       // yes, get the current frequency (in khz)
        temp = temp + (encoder*25);       // add the encoder to it. (in 25Khz steps)
        temp = constrain(temp, MinFreq, MaxFreq); // limit the frequency to valid band
        (RXFrequs[channel]) = temp;       // update the channel frequency
        DisplayFreq = (RXFrequs[channel]);// get the current channel frequency
        ShowFreqAdj();                    // display the frequency
      } 
    }                                     // go back to do it again if mode still valid 
    DisplayFreq = RXFrequs[channel];      // finally, display the frequency in normal text.
    ShowFreq();  
}
